
#include "palmdep.h"

#include "shared.h"
#include "hvc.h"

/* Mark a pattern as dirty */
#define MARK_BG_DIRTY(addr)                                \
{                                                          \
    int name = (addr >> 5) & 0x1FF;                        \
    if(bg_name_dirty[name] == 0)                           \
    {                                                      \
        bg_name_list[bg_list_index] = name;                \
        bg_list_index += 1;                                \
    }                                                      \
    bg_name_dirty[name] |= (1 << ((addr >> 2) & 7));       \
}


/* VDP context */
t_vdp vdp;


/* Initialize VDP emulation */
void vdp_init(void)
{
    vdp_reset();
}


/* Reset VDP emulation */
void vdp_reset(void)
{
    memset(&vdp, 0, sizeof(t_vdp));
    vdp.limit = 1;
}

void vdp_reg_w(uint8 r, uint8 d)
{
    /* Store register data */
    vdp.reg[r] = d;

    switch(r)
    {
        case 0x00: /* Mode Control No. 1 */
            if(vdp.hint_pending)
            {
                if(d & 0x10)
                    z80_set_irq_line(0, ASSERT_LINE);
                else
                    z80_set_irq_line(0, CLEAR_LINE);
            }
            break;

        case 0x01: /* Mode Control No. 2 */
            if(vdp.vint_pending)
            {
                if(d & 0x20)
                    z80_set_irq_line(0, ASSERT_LINE);
                else
                    z80_set_irq_line(0, CLEAR_LINE);
            }
            break;

        case 0x02: /* Name Table A Base Address */
            vdp.ntab = (vdp.reg[2] << 10) & 0x3800;
            break;

        case 0x05: /* Sprite Attribute Table Base Address */
            vdp.satb = (vdp.reg[5] << 7) & 0x3F00;
            break;
    }
}


/* Write data to the VDP's control port */
void vdp_ctrl_w(int data)
{
    /* Waiting for the reset of the command? */
    if(vdp.pending == 0)
    {
        /* Save data for later */
        vdp.latch = data;

        /* Set pending flag */
        vdp.pending = 1;
    }
    else
    {
        /* Clear pending flag */
        vdp.pending = 0;

        /* Extract code bits */
        vdp.code = (data >> 6) & 3;

        /* Make address */
        vdp.addr = (data << 8 | vdp.latch) & 0x3FFF;

        /* Read VRAM for code 0 */
        if(vdp.code == 0)
        {
            /* Load buffer with current VRAM byte */
            vdp.buffer = vdp.vram[vdp.addr & 0x3FFF];

            /* Bump address */
            vdp.addr = (vdp.addr + 1) & 0x3FFF;
        }

        /* VDP register write */
        if(vdp.code == 2)
        {
            int r = (data & 0x0F);
            int d = vdp.latch;

            vdp_reg_w(r, d);
        }
    }
}


/* Read the status flags */
int vdp_ctrl_r(void)
{
    /* Save the status flags */
    uint8 temp = vdp.status;

    /* Clear pending flag */
    vdp.pending = 0;

    /* Clear status flags */
    vdp.status = 0;

    vdp.vint_pending = 0;
    vdp.hint_pending = 0;
    z80_set_irq_line(0, CLEAR_LINE);

    /* Return the old status flags */
    return temp;
}


/* Write data to the VDP's data port */
void vdp_data_w(int data)
{
    int index;

    /* Clear the pending flag */
    vdp.pending = 0;

    switch(vdp.code)
    {
        case 0: /* VRAM write */
        case 1: /* VRAM write */
        case 2: /* VRAM write */

            /* Get current address in VRAM */
            index = (vdp.addr & 0x3FFF);

            /* Only update if data is new */
            if(data != vdp.vram[index])
            {
                /* Store VRAM byte */
                vdp.vram[index] = data;

                MARK_BG_DIRTY(vdp.addr);

//              /* Mark patterns as dirty */
//              vram_dirty[(index >> 5) & 0x01FF] = is_vram_dirty = 1;
            }
            break;

        case 3: /* CRAM write */
            if(cart.type == TYPE_GG)
            {
                if(vdp.addr & 1)
                {                    
                    vdp.cram_latch = (vdp.cram_latch & 0x00FF) | ((data & 0xFF) << 8);
                    vdp.cram[(vdp.addr & 0x3E) | (0)] = (vdp.cram_latch >> 0) & 0xFF;
                    vdp.cram[(vdp.addr & 0x3E) | (1)] = (vdp.cram_latch >> 8) & 0xFF;
                    palette_sync((vdp.addr >> 1) & 0x1F);
                }
                else
                {
                    vdp.cram_latch = (vdp.cram_latch & 0xFF00) | ((data & 0xFF) << 0);
                }
            }
            else
            {
                index = (vdp.addr & 0x1F);
                if(data != vdp.cram[index])
                {
                    vdp.cram[index] = data;
                    palette_sync(index);
                }
            }
            break;
    }

    /* Bump the VRAM address */
    vdp.addr = (vdp.addr + 1) & 0x3FFF;
}


/* Read data from the VDP's data port */
int vdp_data_r(void)
{
    uint8 temp = 0;
    vdp.pending = 0;
    temp = vdp.buffer;
    vdp.buffer = vdp.vram[vdp.addr & 0x3FFF];
    vdp.addr = (vdp.addr + 1) & 0x3FFF;
    return temp;
}

uint8 vdp_vcounter_r(void)
{
    return vc_table[0][vdp.line & 0x1FF];
}


uint8 vdp_hcounter_r(void)
{
    int pixel = (((z80_ICount % CYCLES_PER_LINE) / 4) * 3) * 2;
    return hc_table[0][(pixel >> 1) & 0x1FF];
}
